//
// THIS CODE IS AUTO-GENERATED BY THE MOORE.IO CLIENT.  ANY CHANGES MADE WILL BE ERASED UPON THE NEXT BUILD.
//


// Defines{% for define in defines %}
{% if define.boolean %}-d {{ define.name }}
{% else %}-d {{ define.name }}={{ define.value }}{% endif %}{% endfor %}


// Directories{% for dir in dirs %}
-i {{ dir }}{% endfor %}


// Files{% for file in files %}
{{ file }}{% endfor %}
